Attribute VB_Name = "modIntelliSense"
'Type for GetCaretPos that returns the x/y coordinates
'of the caret (in this case)
Private Type POINTAPI
    X As Long
    Y As Long
End Type

'Returns the caret position in a textbox using
'pixel coordinates
Private Declare Function GetCaretPos Lib "user32" (lpPoint As POINTAPI) As Long

'An enum used to determine what type of viewing
'style the user wants
Public Enum vhIntelliType
  vhViewTagLst = 0
  vhViewAllAttribLst = 1
End Enum

'Tells if the user is in the tag or attrib section
'while typing in the textbox
'It fires only when the scoping rules are correct
Public tagFlag As Boolean
Public attribFlag As Boolean

'Const telling where the tag icon is at on the list
Private Const TAGICON As Byte = 38

Public AttribLst() As String
Public AttribLstCount As Byte

Public AttribTypeLst() As String
Public AttribTypeLstCount As Byte

'Contains a list of attributes, the icon #
'associated w/ them, and the total # of attributes
'loaded
Public AttribNames() As String
Public AttribNameIndex() As Byte
Public AttribCount As Byte

'Contains a list of tag names, their index,
'their list of attributes, and description
Public TagNames() As String
Public TagAttribList() As String
Public TagAttribType() As String
Public TagDesc() As String
Public TagCount As Integer

'Contains the current tag you're in
Public CurrTag As String
'Sets what type of viewing the user would like...
'Do they want to see all the attributes for every
'tag or do they want to see just those ones
'defined for the tag they're in???
Public CurrIntelliViewStyle As vhIntelliType


'Loads the list of tags, etc., described above
Public Sub LoadTagMap(TagMapFile As String)
Dim cToken As String, cLn As String
Dim bef As String, aft As String
Dim i As Integer
Dim f As Byte

'What separator we're looking for
cToken = "="
'Load a free file into f variable
f = FreeFile
'Close any open files
Close
'Make sure counter's set properly before starting
'It's 1 on this b/c # isn't added till the end of
'the first set of tag info.
i = 1
'Open up attribute map and add entries
Open TagMapFile For Input As #f
  On Error Resume Next
  Do While Not EOF(f)
    'Read a line at a time from the file
    Line Input #f, cLn
    
    'Check to make sure it isn't part of the info.
    'used to load the tag list
    If Left(cLn, 1) <> "[" And cLn <> "" Then
    
    If Left(cLn, 1) = "<" Then 'It's a tag
         ReDim Preserve TagNames(i) As String
         'Adds to array and takes off <>s
         TagNames(i) = Mid(cLn, 2, Len(cLn) - 2)
       Else
         'bef is attrib or desc
         'aft is attriblist or actual description
         bef = GetTokenTxt(cLn, cToken, True)
         aft = GetTokenTxt(cLn, cToken, False)
         
         'Removes semi-colon from the end of the line
         aft = Mid(aft, 1, Len(aft) - 1)
         'See what we're on and set the appropriate
         'array info.
         'Reorganize the array and keep all previously
         'added entries (don't want them deleted)
         Select Case LCase(bef)
           Case "attrib"
             ReDim Preserve TagAttribList(i) As String
             TagAttribList(i) = aft
           Case "desc"
             ReDim Preserve TagDesc(i) As String
             TagDesc(i) = aft
           Case "attribtype"
             ReDim Preserve TagAttribType(i) As String
             TagAttribType(i) = aft
         End Select
         
         'Keep a counter going if it's end of info.
         If LCase(bef) = "desc" Then i = i + 1
    End If
    
    End If
  Loop
Close #f

'i contains the total # of tags in the array
'It's i-1 b/c it adds an extra one at the end
TagCount = i - 1

End Sub


'Loads the list of attributes, etc., described above
Public Sub LoadAttribMap(AttribMapFile As String)
Dim cToken As String, cLn As String
Dim bef As String, aft As String
Dim i As Integer
Dim f As Byte

'What separator we're looking for
cToken = "="
'Load a free file into f variable
f = FreeFile
'Close any open files
Close
'Make sure counter's set properly before starting
i = 0
'Open up attribute map and add entries
Open AttribMapFile For Input As #f
  On Error Resume Next
  Do While Not EOF(f)
    'Read a line at a time from the file
    Line Input #f, cLn
    If Left(cLn, 1) <> ";" Then
    'Keep a counter going
    i = i + 1
    
    'bef is the attribute
    'aft is the attribute name index
    bef = GetTokenTxt(cLn, cToken, True)
    aft = GetTokenTxt(cLn, cToken, False)
    
    'Reorganize the array and keep all previously
    'added entries (don't want them deleted)
    ReDim Preserve AttribNames(i) As String
    ReDim Preserve AttribNameIndex(i) As Byte
    
    'Insert info. to the array
    'Make sure aft <= 255 b/c it's declared as byte
    AttribNames(i) = Trim(LCase(bef))
    AttribNameIndex(i) = Val(aft)
    End If
  Loop
Close #f

'i contains the total # of attributes in the array
AttribCount = i
End Sub

'Returns the text before a certain token
'Example:
'  GetTokenTxt "Hello=There","=",True
'    returns "Hello"
'  GetTokenTxt "Hello=There","=",False
'    returns "There"
Public Function GetTokenTxt(s As String, t As String, isBefore As Boolean) As String
Dim c As Integer
Dim cRet As String

c = InStr(1, s, t)
If c <> 0 Then 'found token
   If isBefore = True Then
        'Return the text before the token
        GetTokenTxt = Mid(s, 1, c - 1)
      Else
        'Return the text after the token
        GetTokenTxt = Mid(s, c + 1)
   End If
End If
End Function

Public Sub SetCaretIntelli(p As PictureBox, l As ListView, r As HSRichEdit)
Dim lTop As Long, offset As Long
Dim X As Long, Y As Long
Dim c As POINTAPI
'Get the caret position in the textbox
Call GetCaretPos(c)

On Error Resume Next
'Convert coordinates to twips to place lstview
X = c.X * Screen.TwipsPerPixelX
Y = c.Y * Screen.TwipsPerPixelY
'Move the lstview according to its screen position
'+ 100 b/c of the 3D border
offset = l.ListItems(1).Height + 100
'Adjust position to make sure it's visible
On Error Resume Next
If Y + l.Height + 100 > r.Top + r.Height Then
     p.Move X, Y - (l.Height) + 100
   Else
     p.Move X, Y + (offset)
End If

'IntelliSense has now been positioned properly
End Sub

'Checks to see if IntelliSense can be viewed
'That is, is the caret position in a tag or
'attribute or not?
'If it is, then the current tag is set and whether
'or not it's in a tag or attribute
Public Function isIntelliViewable(r As HSRichEdit, Optional isTypedIn As Boolean = False) As Boolean
Dim fPos As Long, sPos As Long, sStart As Long
Dim sSpaceBef As Long, sBegBef As Long
Dim cBeg As String, cEnd As String
Dim cTxt As String, cTag As String
Dim cPosInTag As Long, cTagBefWithSpace As Long
Dim cTagBefPos As String
Dim c As Boolean

cTxt = r.Text
sStart = r.SelStart + 1
cBeg = "<"
cEnd = ">"

'Make sure to set public variables back to normal
CurrTag = ""
tagFlag = False
attribFlag = False
On Error Resume Next
With r
  'Find position of the next < and >
  fPos = InStr(sStart, cTxt, cBeg)
  sPos = InStr(sStart, cTxt, cEnd)
  
  'The user's typing in the tag...
  If isTypedIn = True Then
     sPos = sStart + 1
  End If
  'If < comes before > then you're not in a tag
  If fPos > sPos Then
       'Set this variable b/c at the end it's used
       'to return the function's value
       c = True 'It is in a tag
       
       sBegBef = InStrRev(sStart, cTxt, "<")
       
       If isTypedIn = True Then
            sBegEnd = sPos + 1
          Else
            sBegEnd = InStr(sBegBef, cTxt, ">")
       End If
       cTag = Mid(cTxt, sBegBef, sBegEnd - sBegBef + 1)
       
       'Remember, InStrRev returns how far it is
       'from the BEGINNING, so we have to think
       'differently
       cPosInTag = sStart - sBegBef
       
       'The text in JUST the tag up to the SelStart
       cTagBefPos = Mid(cTag, 1, cPosInTag)
       'The text up to the space
       cTagBefWithSpace = InStr(1, cTagBefPos, " ")
       'If there's a space in the text up to the pos
       'you're in, then you know you're in an
       'attribute and not a tag
       'Sets the public variables
       'so we know which tag it's in
       'and if it's in a tag or attribute
       
       

       
       
       If cTagBefWithSpace <> 0 Then
            'Set the current tag and it is in an
            'attribute
            CurrTag = Trim(Mid(cTagBefPos, 2, cTagBefWithSpace - 2))
            tagFlag = False
            attribFlag = True
          Else
            'Set the current tag and it is in a
            'tag
            'Uses cTag b/c in this case, we don't
            'know the entire tag, we just know
            'we're in a tag, if we used cTagBefPos
            'If we use cTag, we can get the tag
            'b/c that contains the entire tag
            'we're in
            CurrTag = Trim(Mid(cTag, 2, InStr(2, cTag, " ") - 2)) ', Len(cTagBefPos))
            If CurrTag = "" Then CurrTag = Mid(cTag, 2, Len(cTag) - 2)
            attribFlag = False
            tagFlag = True
       End If
     Else
       c = False
  End If
End With

'Check to see if it's a comment b/c those don't
'follow the rules
If Left(CurrTag, 1) = "!" Then CurrTag = "!--"

If isTypedIn = True Then
'Make sure there's no errors if typed in...
If InStr(1, CurrTag, ">") <> 0 Then CurrTag = ""
If Right(CurrTag, 2) = vbCrLf Then CurrTag = Left(CurrTag, Len(CurrTag) - 2)
End If
'Get rid of any vbCrLf's in the tag...
CurrTag = ClearTagOfReturns(CurrTag)

'CurrTag now holds the current tag you're in
'tagFlag and attribFlag says if you're in a tag
'or attribute
isIntelliViewable = c
End Function

'InStr in reverse -- returns the position of a string
'before wherever you're at
'The return info. is from the BEGINNING so watch
'out!!!  (c:
Public Function InStrRev(cStart As Long, cSrchTxt As String, cSrch As String) As Long
Dim i As Long

'Cycle through and find it's position from the left
For i = 1 To cStart
  i = InStr(i, cSrchTxt, cSrch)
  If InStr(i + Len(cSrch), cSrchTxt, cSrch) > cStart Then
     'Position is found, exit loop
     'The position is i
     Exit For
  End If
Next i

'It can't be the same as InStr b/c it's backwards
If i = InStr(cStart + Len(cSrch), cSrchTxt, cSrch) Then
   i = 0
End If

'Return position
InStrRev = i
End Function

'Clears any listview
Public Sub ClearLstView(l As ListView)
  l.ListItems.Clear
End Sub

'Loads the IntelliSense info. and returns its
'description
Public Function LoadIntelliSense(r As HSRichEdit, l As ListView, p As PictureBox, Optional KeyAscii As Integer = 0) As String
Static currBool As Boolean
Dim i As Integer

If CurrTag = "!--" Then Exit Function

'Check to see if you've typed a tag in...
If KeyAscii <> 0 Then
   If KeyAscii = 60 Then 'You're beginning a tag
        currBool = True
      ElseIf KeyAscii = 62 Then 'Ending a tag
        CurrTag = ""
        attribFlag = False
        tagFlag = False
        currBool = False
      ElseIf KeyAscii = 32 And currBool = True Then
        tagFlag = False
        attribFlag = True
   End If
End If

'You're not in a tag, so get outta here
If isIntelliViewable(r, currBool) = False Then
   p.Visible = False
   l.Visible = p.Visible
   Exit Function
End If
'You're in a tag now...load the listview...

'Clear all the items from the listview
Call ClearLstView(l)
'Load the lstview appropriately...
'CurrIntelliViewStyle is the public variable
'set by the user
p.Visible = LoadLstView(l, CurrIntelliViewStyle)
l.Visible = p.Visible
'Lstview is now loaded correctly

'Find tag index to get description...
For i = 1 To TagCount
  If LCase(CurrTag) = LCase(TagNames(i)) Then
     Exit For
  End If
Next i

'Set the listview to the appropriate place
Call SetCaretIntelli(p, l, r)

'i now contains the tag index...
On Error Resume Next
'Returns the description of the tag...
LoadIntelliSense = TagDesc(i)


End Function

'Returns a string containing a general outline
'of the tag the user's on...for example...
'if the user were in a <body> tag...it might
'return something like:
'<body bgcolor=color vlink=color bgsound=file>
'etc., etc.
Public Function LoadTagOptions(Optional pos As Integer) As String
Dim cAttribLst As String, cAttribTypeLst As String
Dim i As Integer
Dim numAttribs As Byte
Dim cType As String, cOpt As String

'CurrTag holds the tag we're in...
'TagAttribList() contains the attrib list associated
'w/ each tag and TagAttribType() contains the
'type associated w/ each attribute...

'pos contains the position in the array for all
'the arrays associated w/ the tag...
If pos = 0 Then
   For i = 1 To TagCount
     If LCase(CurrTag) = LCase(TagNames(i)) Then Exit For
   Next i
   
   pos = i
End If
'Gets the attribute and type list...
cAttribLst = TagAttribList(pos) & ","
cAttribTypeLst = TagAttribType(pos) & ","

'Gets the # of attributes in the list...
numAttribs = GetNumTokensInStr(cAttribLst)

'If there's not enough attributes, then don't
'do the loop...
If numAttribs > 1 Then
     For i = 1 To numAttribs
       cType = FieldInString(cAttribTypeLst, i)
       cOpt = cOpt & " " & UCase(FieldInString(cAttribLst, i)) & "=" & LCase(cType)
     Next i

     cOpt = "<" & UCase(CurrTag) & " " & Trim(cOpt) & ">"
   Else
     cOpt = "<" & UCase(Trim(CurrTag)) & ">"
End If

'Return the general tag info...
LoadTagOptions = cOpt
'MsgBox (cAttribLst & vbCrLf & cAttribTypeLst & vbCrLf & numAttribs)
End Function

Public Function ClearTagOfReturns(s As String) As String
Dim cPos

cPos = InStr(1, s, vbCrLf)
If cPos <> 0 Then
     ClearTagOfReturns = Mid(s, 1, cPos - 1)
   Else
     ClearTagOfReturns = s
End If
End Function

'Returns visibile back to LoadIntelliSense
Public Function LoadLstView(l As ListView, c As vhIntelliType) As Boolean
Dim i As Integer
Dim cIndex As Integer

'Reset variable
ReDim AttribLst(0)
'Make sure everything's cleared...
Call ClearLstView(l)

'If it's a tag, then load tag list...
On Error Resume Next
If tagFlag = True Then
  For i = 1 To TagCount
    With l
      .ListItems.Add , , TagNames(i), TAGICON, TAGICON
    End With
Next i
End If

'If it's an attribute, load the list according
'to the user options...
If attribFlag = True Then
   'User has selected to view only the attributes
   'associated with the tag
   If c = vhViewTagLst Then
     Dim cAttribLst As String
     Dim cTypeLst As String
     
     'Find the index of the tag so we can get the
     'attribute list...
     For i = 1 To TagCount
       If LCase(CurrTag) = LCase(TagNames(i)) Then
          cIndex = i
          Exit For
       End If
     Next i
     'Get the attribute list based on the tag...
     
     On Error Resume Next
     cAttribLst = TagAttribList(cIndex) & ","
     cTypeLst = TagAttribType(cIndex) & ","
     
     'All the lists are now loaded...
     'The information we now have is:
     'AttribLst() = array of attributes for tag
     'AttribTypeLst() = array of attrib types for
     '  each attribute
     Call LoadAttributeLst(cAttribLst)
     Call LoadAttributeTypeLst(cTypeLst)
     
     Dim a As Byte
     Dim cLV As ListItem
     
     For i = 1 To AttribLstCount
       For a = 1 To AttribCount
         If LCase(AttribLst(i)) = LCase(AttribNames(a)) Then
            'Get the icon index from the mapped
            'attribute icon list
            cIndex = AttribNameIndex(a)
            Exit For
         End If
         cIndex = 0
       Next a
       
       'If there's no associated icon, then add a
       'general purpose icon, which is #11 in the
       'imagelist
       If cIndex = 0 Then cIndex = 11
       
       'Add the item...
       On Error Resume Next
       Set cLV = l.ListItems.Add(, , AttribLst(i), cIndex, cIndex)
       cLV.Tag = AttribTypeLst(i)
     Next i
     
   End If
   
   'User wants to see all the attributes in the list
   If c = vhViewAllAttribLst Then
      'Add the attribute list to the listview
      'Note:  The attribute type isn't set b/c
      'it's type is set by the tag...the same
      'attribute may have different types depending
      'on the tag...
      For i = 1 To AttribCount
        With l
          'cLV is now the listview added...
          Set cLV = .ListItems.Add(, , AttribNames(i), AttribNameIndex(i), AttribNameIndex(i))
        End With
      Next i
   End If
End If

'Make a final check...
On Error Resume Next
If l.ListItems.Count <= 1 And Trim(l.ListItems(1).Text) = "" Then
     Call ClearLstView(l)
     LoadLstView = False
   
   Else
     LoadLstView = True
End If
End Function

Public Sub LoadAttributeLst(sAttribLst As String)
Dim i As Byte

     AttribLstCount = 0
     'Count the # of ,'s are in the list...
     For i = 1 To Len(sAttribLst)
         i = InStr(i, sAttribLst, ",")
         'If there's no more, then exit!!
         If i = 0 Then
            Exit For
         End If
         
         'Add 1 to list...
         AttribLstCount = AttribLstCount + 1
     Next i
     
     ReDim Preserve AttribLst(AttribLstCount) As String
     For i = 1 To AttribLstCount
       AttribLst(i) = FieldInString(sAttribLst, i)
     Next i
End Sub

Public Sub LoadAttributeTypeLst(sAttribTypeLst As String)
Dim i As Byte

     AttribTypeLstCount = 0
     'Count the # of ,'s are in the list...
     For i = 1 To Len(sAttribTypeLst)
         i = InStr(i, sAttribTypeLst, ",")
         'If there's no more, then exit!!
         If i = 0 Then
            Exit For
         End If
         
         'Add 1 to list...
         AttribTypeLstCount = AttribTypeLstCount + 1
     Next i
     
     ReDim Preserve AttribTypeLst(AttribTypeLstCount) As String
     For i = 1 To AttribTypeLstCount
       AttribTypeLst(i) = FieldInString(sAttribTypeLst, i)
     Next i
End Sub

Private Function FieldInString(ByVal StringArg As String, ByVal FieldNumber As Integer) As String

    ' This Function returns requested field # from str
    '     ing argument
    Dim X, Y, oldY As Integer
    'add a delimiting space to the end of the argument
    '
    StringArg = StringArg + " "
    X = 0
    Y = 1
    oldY = 1

    While Y < Len(StringArg) And X < FieldNumber
        Y = InStr(oldY, StringArg, ",")
        On Error Resume Next
        FieldInString = Mid$(StringArg, oldY, Y - oldY)
        X = X + 1
        oldY = Y + 1
    Wend

        'If requested field # was greater than the # of wo
        '     rds in
        'passed string argument, then return an empty stri
        '     ng

        If X < FieldNumber Then FieldInString = ""
End Function

Public Function GetNumTokensInStr(s As String, Optional Token As String = ",") As Byte
Dim i As Byte
Dim AttribLstCount As Byte

     AttribLstCount = 0
     'Count the # of ,'s are in the list...
     For i = 1 To Len(s)
         i = InStr(i, s, Token)
         'If there's no more, then exit!!
         If i = 0 Then
            Exit For
         End If
         
         'Add 1 to list...
         AttribLstCount = AttribLstCount + 1
     Next i
     
GetNumTokensInStr = AttribLstCount
End Function
